// Class implementation for ParSledExecutor
//
// Copyright Matthias Gaertner, Schenefeld, Germany, 2001
// You may use this code under the terms of the GPL.
// Updated packages may be found at http://www.rtner.de/parsled/
//

#include <stdio.h>
#include <ParSledExecutor.h>
#include <ConfigGroup.h>
#include <ConfigLed.h>

#ifndef SIMULATION
#include <sys/ioctl.h>
#include <fcntl.h>
#include <stdlib.h>
#include <unistd.h>
#include <utmp.h>
#include <linux/lp.h>
#endif

ParSledExecutor::ParSledExecutor() :
    _countButtonPressed( 0 ),
    _fd( -1 ),
    _sc( NULL )
{
#ifdef SIMULATION
    _press = 0;
#else
    _lstate = 0;
#endif
}

ParSledExecutor::~ParSledExecutor()
{
#ifndef SIMULATION
    if( _fd >= 0 )
    {
        // Reset LEDs
        _lstate = 0;
        switch( _interfacetype )
        {
            case serial:
                ioctl( _fd, TIOCMSET, &_lstate );
                break;

            case parport:
                // write only if !busy
                int tmpstate;
                ioctl( _fd, LPGETSTATUS, &tmpstate );
                if( tmpstate & LP_PBUSY )
                {
                    ioctl( _fd, LPRESET );
                    write( _fd, &_lstate, 1 );
                }
                break;

            default:
                // error
                break;
        }
        close( _fd );
    }
#endif
}

int ParSledExecutor::start( ParSledConfiguration *sc )
{
    int r = 0;

    _sc = sc;
    _interfacetype = serial;  // default

#ifndef SIMULATION
    if( (_fd=open( _sc->getGeneral()->getDevice(), O_RDWR | O_NDELAY)) < 0 )
    {
        char buf[128];
        snprintf(buf, sizeof(buf)-1, "Open device %s: ", _sc->getGeneral()->getDevice());
        perror(buf);
        return 404;   // this becomes a 148 because 404 % 256 = 148
    }
    else
    {
        // try to find out which type of interface we use
        _lstate = 0;
        // 1st check serial port
        if( ioctl( _fd, TIOCMSET, &_lstate ) == 0 )
        {
            _interfacetype = serial;
        }
        else if( ioctl( _fd, LPGETSTATUS, &_lstate ) == 0 )
        {
            // 2nd try: parallel port
            _lstate = 4; // set bit2 on, otherwise bit1 is too weak to drive the LED
            _interfacetype = parport;
            ioctl( _fd, LPRESET );
        }
        else
        {
            perror("Check parallel port");
        }
    }
#endif

    // Initialize LED status etc.
    if( (r=_handleLEDs()) != 0 )
    {
        return r;
    }
    return r;
}

int ParSledExecutor::execute( void )
{
    int r = 0;
    for(;;)
    {
        //
        // Part one: detect and handle button
        //
        if( _buttonPressed() != 0 )
        {
            _countButtonPressed++;
        }
        else if( _countButtonPressed == 0 )
        {
            // Was not pressed, is not pressed: do nothing here
        }
        else
        {
            // Action!
            if( (r=_buttonPressDetected()) != 0 )
            {
                break;
            }
        }

        //
        // Part two: handle LED displays
        //
        if( (r=_handleLEDs()) != 0 )
        {
            break;
        }

        //
        // Part three: wait
        //

#ifdef SIMULATION
        char a[1000];
        a[0] = '_';
        gets( a );
        if( a[0] == 'x' )
        {
            break;
        }
        if( a[0] == ' ' )
        {
            _press = !_press;
        }
#else
        usleep( _sc->getGeneral()->getInterval() * 1000 );  // config file setting in milliseconds
#endif
    }
    return r;
}

int ParSledExecutor::stop()
{
    return 0;
}

int ParSledExecutor::_buttonPressed( void )
{
    int r = 0;
#ifdef SIMULATION
    r = _press;
#else
    int state = 0;
    switch( _interfacetype )
    {
        case serial:
            {
            // Due to the electric design it is not possible to read the push
            // button status when the green LED is off. The button appears not
            // to be pressed in this case.
            // Solution: briefly switch the green LED on, read the button, then
            // switch it back to its original state. This happens too fast for
            // the human eye to notice.
            //
            // Caveat: Is the above description really correct?
            // I am not sure, but the code works.
            int tmp_lstate = _lstate | TIOCM_DTR;
            ioctl( _fd, TIOCMSET, &tmp_lstate );

            ioctl( _fd, TIOCMGET, &state ); /* read interface */
            r = (state & TIOCM_CAR) ? 1 : 0;

            ioctl( _fd, TIOCMSET, &_lstate );
            }
            break;

        case parport:
            // parallel port (select signal)
            ioctl( _fd, LPGETSTATUS, &state );    // read interface
            r = (state & LP_PSELECD) ? 0 : 1;
            break;

        default:
            // error
            break;
        }
#endif
    return r;
}

int ParSledExecutor::_execute( ConfigAction *ca )
{
#ifdef SIMULATION
    printf( "Execute: %s\n", ca->getCommand().getPointerC() );
#else
    system( ca->getCommand().getPointerC() );
#endif
    return 0;
}

int ParSledExecutor::_buttonPressDetected( void )
{
    unsigned int n = _countButtonPressed;
    _countButtonPressed = 0;

    // Look at the list of groups and see which one needs to be executed.
    ParSledConfigClass *scc;
    ConfigAction *ca;
    _sc->resetIter();
    while( (scc=(ParSledConfigClass*)_sc->getNextIter()) != NULL )
    {
        if( scc->getClassName().equals("GROUP") )
        {
            ConfigGroup *cg = (ConfigGroup*) scc;
            if( cg->getMin() <= n && n <= cg->getMax() )
            {
                unsigned int groupState = cg->getState();

                cg->resetIter();
                while( (ca=(ConfigAction*)cg->getNextIter()) != NULL )
                {
                    if( groupState == ca->getState() )
                    {
                        _execute( ca );
                        cg->setState( ca->getNext() );
                        break;
                    }
                }
                break;
            }
        }
    }
    return 0;
}

int ParSledExecutor::_setLED( int n, bool on_off )
{
#ifdef SIMULATION
    if( n == 0 )
    {
        printf( "\r%s ", (on_off) ? "on " : "off" );
    }
    else
    {
        printf( "%s ", (on_off) ? "on " : "off" );
    }
#else
    int bitpat = 0;
    int tmpstate = 0;

    switch( _interfacetype )
    {
        case serial:
            if( n == 0 )      // yellow LED in original design
            {
                bitpat = TIOCM_RTS;
            }
            else if( n == 1 ) // green LED in original design
            {
                bitpat = TIOCM_DTR;
            }
            break;

        case parport:
            bitpat = 1 << n;
            break;

        default:
            // error
            break;
    }

    if( on_off )
    {
        _lstate |= bitpat;
    }
    else
    {
        _lstate &= ~bitpat;
    }

    switch( _interfacetype )
    {
        case serial:
            ioctl( _fd, TIOCMSET, &_lstate );
            break;

        case parport:
            // write only if !busy
            ioctl( _fd, LPGETSTATUS, &tmpstate );
            if( tmpstate & LP_PBUSY )
            {
                ioctl( _fd, LPRESET );
                write( _fd, &_lstate, 1 );
            }
        break;

        default:
            // error
            break;
    }
#endif
    return 0;
}

int ParSledExecutor::_handleLEDs( void )
{
    ParSledConfigClass *cg;
    ConfigLed *cl;
    _sc->resetIter();
    while( (cg=(ParSledConfigClass*)_sc->getNextIter()) != NULL )
    {
        if( cg->getClassName().isEmpty() )
        {
            cg->resetIter();
            while( (cl=(ConfigLed*)cg->getNextIter()) != NULL )
            {
                _handleLED( cl );
            }
            break;
        }
    }
    return 0;
}

int ParSledExecutor::_handleLED( ConfigLed *cl )
{
    unsigned int m = cl->getMode();
    unsigned int c = cl->getCycle();
    int f = cl->getFlashNumber();

    if( m == 0 )        // BLINK
    {
        bool b = (c < cl->getDutyCycleOn());

        c++;
        if( c >= cl->getDutyCycleOn() + cl->getDutyCycleOff() )
        {
            c = 0;
        }
        cl->setCycle(c);
        _setLED( cl->getPhysical(), b );
    }
    else if( m == 1 )   // ON
    {
        _setLED( cl->getPhysical(), true );
    }
    else if( m == 2 )   // OFF
    {
        _setLED( cl->getPhysical(), false );
    }
    else if( m == 3 )   // BUTTON
    {
        _setLED( cl->getPhysical(), (_countButtonPressed > 0) );
    }
    else if( m == 4 )   // COMMAND
    {
        // light LED when a command could be executed by releasing the button
        unsigned int n = _countButtonPressed;
        bool b = false;

        // Look at the list of groups and see which one could be executed.
        ParSledConfigClass *scc;
        ConfigAction *ca;
        _sc->pushIter();
        while( n > 0 && (scc=(ParSledConfigClass*)_sc->getNextIter()) != NULL )
        {
            if( scc->getClassName().equals("GROUP") )
            {
                ConfigGroup *cg = (ConfigGroup*) scc;
                if( cg->getMin() <= n && n <= cg->getMax() )
                {
                    unsigned int groupState = cg->getState();

                    cg->resetIter();
                    while( (ca=(ConfigAction*)cg->getNextIter()) != NULL )
                    {
                        if( groupState == ca->getState() )
                        {
                            b = true;
                            break;
                        }
                    }
                    break;
                }
            }
        }
        _sc->popIter();
        _setLED( cl->getPhysical(), b );
    }
    else if( m == 5 )   // RUNLEVEL
    {
        // Flash runlevel
        if( c > 0 && f > 0 )
        {
            c--;    // Wait
        }
        else
        {
            c = cl->getDutyCycleOn()-1;

            if( f > 0 )
            {
                bool b = ((f%2)==0);
                f--;
                _setLED( cl->getPhysical(), b );
            }
            else if( f == 0 )
            {
                f = - ((int)cl->getDutyCycleOff()+1);
                _setLED( cl->getPhysical(), false );
            }
            else if( f < -1 )
            {
                f++;
            }
            else if( f == -1 )
            {
                f = _getRunlevel() * 2;
            }
        }

        cl->setCycle(c);
        cl->setFlashNumber(f);
    }
    return 0;
}

int ParSledExecutor::_getRunlevel( void )
{
    int rl = 2;
#ifndef SIMULATION
    struct utmp *ut;

    setutent();
    while( (ut = getutent()) != NULL )
    {
        if (ut->ut_type == RUN_LVL)
        {
            rl = (ut->ut_pid % 256);
            break;
        }
    }
    endutent();
#endif
    return rl;
}

/* */
